'' =================================================================================================
''
''   File....... jm_ds3231.spin
''   Purpose.... DS3231 RTC interface (uses 24-hr format)
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (c) 2014-2025 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon.mcphalen@gmail.com
''   Started....
''   Updated.... 06 APR 2025
''
'' =================================================================================================


con

  VERSION  = 1_1_0                                              ' added utilities


con { fixed io pins }

  RX_PGM   = 31  { I }                                          ' serial / programming
  TX_PGM   = 30  { O }

  SDA1     = 29  { I/O }                                        ' i2c / eeprom
  SCL1     = 28  { I/O }

  LED27    = 27  { O }                                          ' LEDs on PAB and FLiP
  LED26    = 26  { O }


con

  #true,  ON, OFF
  #false, NO, YES


con { i2c }

  ACK = i2c#ACK
  NAK = i2c#NAK

  DS3231_WR = %1101_000_0                                       ' i2c slave id
  DS3231_RD = %1101_000_1


  #$00, R_SCS, R_MNS, R_HRS                                     ' time registers
  #$03, R_DAY, R_DATE, R_MON, R_YR                              ' date registers

  #$07, R_A1SCS, R_A1MNS, R_A1HRS                               ' alarm 1 time registers
  #$0A, R_A1DYDT                                                ' alarm 1 day / date

  #$0B, R_A2MNS, R_A2HRS                                        ' alarm 2 time registers
  #$0D, R_A2DYDT                                                ' alarm 2 day / date

  #$0E, R_CTRL                                                  ' control register
  #$0F, R_STATUS                                                ' status register

  #$10, R_OFFSET                                                ' aging offset

  #$11, R_TMSB, R_TLSB                                          ' temp registers


obj

  i2c : "jm_i2c"


pub null

'' This is not an application


pub start(scl, sda) | check

'' Connects DS3231 to user-defined I2C bus

  i2c.setupx(scl, sda)
  if (present)
    check := rd_reg(R_STATUS)
    if (check & $80)                                            ' clear OSF bit if set
      check &= $7F
      wr_reg(R_STATUS, check)


pub present : result

  return i2c.present(DS3231_WR)


pub set_time(sc, mn, hr)

'' Set seconds, minutes, and hours registers
'' -- values are BCD

  i2c.start
  i2c.write(DS3231_WR)
  i2c.write(R_SCS)
  i2c.write(sc)
  i2c.write(mn)
  i2c.write(hr)
  i2c.stop


pub set_calendar(dy, dt, mo, yr)

'' Set calendar
'' -- values are BCD

  i2c.start
  i2c.write(DS3231_WR)
  i2c.write(R_DAY)
  i2c.write(dy)
  i2c.write(dt)
  i2c.write(mo)
  i2c.write(yr)
  i2c.stop


pub write(n, r, p_src)

'' Write n BCD registers at p_src to DS3231
'' -- r is register number in DS3231
'' -- values at p_src are BCD

  i2c.start
  i2c.write(DS3231_WR)
  i2c.write(r)
  repeat n
    i2c.write(byte[p_src++])
  i2c.stop


pub wr_reg(r, value)

'' Writes specific BCD register in DS3231

  write(1, r, @value)


pub wr_time(p_src)

'' Write time (secs, mins, hours) registers at p_src to DS3231

  write(3, R_SCS, p_src)


pub wr_calendar(p_src)

'' Write calendar (day, date, month, year) registers at p_src to DS3231

  write(4, R_DAY, p_src)


pub read(n, p_dest, r)

'' Reads n regsiters at addr in DS3231 to p_dest
'' -- p_dest points to byte(s)
'' -- r is the beginning register in DS3231

  i2c.start
  i2c.write(DS3231_WR)
  i2c.write(r)                                                  ' set address pointer
  i2c.start                                                     ' restart for read
  i2c.write(DS3231_RD)
  repeat (n-1)
    byte[p_dest++] := i2c.read(ACK)
  byte[p_dest++] := i2c.read(NAK)
  i2c.stop


pub rd_reg(r) : result

'' Reads single registers from DS3231

  read(1, @result, r)


pub rd_time(p_dest)

'' Reads clock registers (secs, mins, hours) from DS3231 to p_dest
'' -- p_dest points to array of bytes

  read(3, p_dest, R_SCS)


pub rd_calendar(p_dest)

'' Reads calendar registers (day, date, month, year) from DS3231 to p_dest
'' -- p_dest points to array of bytes

  read(4, p_dest, R_DAY)


pub time_code : result

'' Returns current time as BCD-formatted long: $00_hh_mm_ss

  read(3, @result, R_SCS)


pub date_code : result

'' Returns current date as BCD-formatted long: $00_yr_mo_dt

  read(3, @result, R_DATE)


con { formatted strings }

pub time_hhmmss(p_src) : result

'' Returns pointer to formatted time string
'' -- formatted string of registers at p_src
'' -- user -1 for internal RTC registers

  if (p_src < 0)
    read(3, @result, R_SCS)                                       ' internal clock
  else
    bytemove(@result, p_src, 3)                                   ' external values

  StrLong[0] := result.byte[2] >>   4 + "0"                       ' convert to ASCII
  StrLong[1] := result.byte[2] &  $0F + "0"
  StrLong[2] := ":"
  StrLong[3] := result.byte[1] >>   4 + "0"
  StrLong[4] := result.byte[1] &  $0F + "0"
  StrLong[5] := ":"
  StrLong[6] := result.byte[0] >>   4 + "0"
  StrLong[7] := result.byte[0] &  $0F + "0"

  return @StrLong


pub time_ampm(p_src) : result | hr

'' Returns pointer to formatted time string
'' -- formatted string of registers at p_src
'' -- user -1 for internal RTC registers

  if (p_src < 0)
    read(3, @result, R_SCS)                                     ' internal clock
  else
    bytemove(@result, p_src, 3)                                 ' external values

  hr := bcd2dec(result.byte[2])                                 ' convert to decimal

  case hr                                                       ' adjust for am/pm
    0..11 :
      TimeStr[9] := "A"
      if (hr == 0)
        hr := 12

    12..23 :
      TimeStr[9] := "P"
      if (hr > 12)
        hr -= 12

  TimeStr[0] := hr  / 10 + "0"                                  ' convert to ASCII
  TimeStr[1] := hr // 10 + "0"
  TimeStr[2] := ":"
  TimeStr[3] := result.byte[1] >>   4 + "0"
  TimeStr[4] := result.byte[1] &  $0F + "0"
  TimeStr[5] := ":"
  TimeStr[6] := result.byte[0] >>   4 + "0"
  TimeStr[7] := result.byte[0] &  $0F + "0"

  return @TimeStr


pub date_mmddyy(p_src) : result

'' Returns pointer to formatted date
'' -- formatted string of registers at p_src
'' -- user -1 for internal RTC registers

  if (p_src < 0)
    read(3, @result, R_DATE)                                    ' internal clock
  else
    bytemove(@result, p_src, 3)                                 ' external values

  StrLong[0] := result.byte[1] >>   4 + "0"
  StrLong[1] := result.byte[1] &  $0F + "0"
  StrLong[2] := "/"
  StrLong[3] := result.byte[0] >>   4 + "0"
  StrLong[4] := result.byte[0] &  $0F + "0"
  StrLong[5] := "/"
  StrLong[6] := result.byte[2] >>   4 + "0"
  StrLong[7] := result.byte[2] &  $0F + "0"

  return @StrLong


pub today : result

'' Return RTC day of week, 1..7 (Sunday..Saturday)

  return rd_reg(R_DAY)


pub day_short(d) : result

'' Return pointer to string with name of day, e.g. "MON"
'' -- use 1..7 to print specific day name
'' -- use -1 for current RTC day

  if (d < 0)
    d := today
  else
    if ((d < 1) or (d > 7))
      return @Unknown

  return @DayShort + (--d << 2)                                 ' 4 chars ea


pub day_long(d) : p_str

'' Return pointer to string with name of day, e.g. "Monday"
'' -- use 1..7 to print specific day name
'' -- use -1 for current RTC day

  if (d < 0)
    d := today
  else
    if ((d < 1) or (d > 7))
      return @Unknown

  p_str := @DayLong                                             ' point to start of names

  repeat --d                                                    ' z-align, 0..6
    p_str += strsize(p_str) + 1                                 ' skip over to target string


pub month_short(m) : p_str

'' Return pointer to string with name of month, e.g. "JAN"
'' -- use $01..$12 (bcd) to print specific month name
'' -- use -1 for current RTC month

  if (m < 0)
    read(1, @m, R_MON)
    m := bcd2dec(m)
  else
    m := bcd2dec(m)
    if (m < 1) or (m > 12)
      return @Unknown

  return @MonthShort + (--m << 2)                                 ' 4 chars ea


pub month_long(m) : p_str

'' Return pointer to string with name of month, e.g. "January"
'' -- use $01..$12 (bcd) to print specific month name
'' -- use -1 for current RTC month

  if (m < 0)
    read(1, @m, R_MON)
    m := bcd2dec(m)
  else
    m := bcd2dec(m)
    if (m < 1) or (m > 12)
      return @Unknown

  p_str := @MonthLong                                           ' point to start of names

  repeat --m                                                    ' z-align, 0..11
    p_str += strsize(p_str) + 1                                 ' skip over to target string


con { utilities }

pub is_leap_year(yr) : result

'' Return true if yr is leap year
'' -- yr is -1 (read RTC), or $00..$99 (BCD)

  if (yr < 0)
    read(1, @yr, R_YR)

  yr := bcd2dec(yr) + 2000                                      ' convert to decimal, full year

  if (yr // 400 == 0)
    return true
  elseif (yr // 100 == 0)
    return false
  elseif (yr // 4 == 0)
    return true
  else
    return false


dat

  DaysInMonth   byte    0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31


pub last_date(mo, yr) : ld                                      ' BCD in, DECIMAL out

'' Return last date for month in given year
'' -- mo is -1 (read RTC), or $01..$12 (BCD)
'' -- yr is -1 (read RTC), or $00..$99 (BCD)
'' -- return is decimal (not BCD)

  if (mo < $00) or (yr < $00)
    read(2, @mo, R_MON)                                         ' read current

  ld := DaysInMonth[bcd2dec(mo)]                                ' read last

  if (mo == $02) and (is_leap_year(yr))                         ' February of leap year?
    ld++


pub ordinal(dt, mo, yr) : od | i                                ' BCD in, DECIMAL out

'' Return Ordinal day (1..366) for date, month, year
'' -- dt is -1 (read RTC), or $01..$31 (BCD)
'' -- mo is -1 (read RTC), or $01..$12 (BCD)
'' -- yr is -1 (read RTC), or $00..$99 (BCD)
'' -- return is decimal (not BCD)

  if (dt < $00) or (mo < $00) or (yr < $00)
    read(3, @dt, R_DATE)                                        ' current date

  od := bcd2dec(dt)                                             ' today in current month
  mo := bcd2dec(mo)

  i := 1
  repeat while (i < mo)                                         ' accumulate previous months
    od += DaysInMonth[i++]

  if (mo > 2) and (is_leap_year(yr))                            ' adjust for leap year
    od += 1


pub julian(dt, mo, yr) : jd                                     ' DECIMAL in and out

'' Returns Julian day (days since Jan 1, 4713 BC)
'' -- dt, mo, and year are decimal input
''    * dt is 1..31
''    * mo is 1..12
''    * yr is -4713..????
'' -- reference: https://pdc.ro.nu/jd-code.html

  yr += 8000

  if (mo < 3)
    yr--
    mo += 12

  jd := (yr*365) + (yr/4) - (yr/100) + (yr/400) - 1200820 + (mo*153+3)/5 - 92 + dt-1


con { conversions }

pub bcd2dec(bcd) : result

'' Return decimal value of 2-digit bcd ($00..$99)

  return ((bcd >> 4) * 10) + (bcd & $0F)


pub dec2bcd(dec) : result

'' Return BCD value of 2-digit 0..99 decimal

  return ((dec / 10) << 4) | (dec // 10)


con { temperature }

pub temp_c : result

'' Returns temperature in degrees Celcius
'' -- returns in 1/100s

  read(1, @result.byte[3], R_TMSB)
  read(1, @result.byte[2], R_TLSB)

  return (result ~> 22) * 25


pub temp_f : result

'' Returns temperature in degrees Fahrenheit
'' -- returns in 1/100s

  return temp_c * 9 / 5 + 32_00


dat { strings }

  Unknown       byte    "???", 0

  StrShort      byte    "00:00", 0
  StrLong       byte    "00:00:00", 0

  TimeStr       byte    "hh:mm:ss xM", 0

  DayShort      byte    "SUN", 0
                byte    "MON", 0
                byte    "TUE", 0
                byte    "WED", 0
                byte    "THU", 0
                byte    "FRI", 0
                byte    "SAT", 0

  DayLong       byte    "Sunday", 0
                byte    "Monday", 0
                byte    "Tuesday", 0
                byte    "Wednesday", 0
                byte    "Thursday", 0
                byte    "Friday", 0
                byte    "Saturday", 0

  MonthShort    byte    "JAN", 0
                byte    "FEB", 0
                byte    "MAR", 0
                byte    "APR", 0
                byte    "MAY", 0
                byte    "JUN", 0
                byte    "JUL", 0
                byte    "AUG", 0
                byte    "SEP", 0
                byte    "OCT", 0
                byte    "NOV", 0
                byte    "DEC", 0

  MonthLong     byte    "January",   0
                byte    "February",  0
                byte    "March",     0
                byte    "April",     0
                byte    "May",       0
                byte    "June",      0
                byte    "July",      0
                byte    "August",    0
                byte    "September", 0
                byte    "October",   0
                byte    "November",  0
                byte    "December",  0


con { license }

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}