'' =================================================================================================
''
''   File........ jm_ds3231_demo2.spin
''   Purpose..... Display RTC values with option of resetting them on start-up
''   Cogs Used...
''   Author...... Jon "JonnyMac" McPhalen
''                Copyright (c) 2025 Jon McPhalen
''                -- see below for terms of use
''   E-mail...... jon.mcphalen@gmail.com
''   Started.....
''   Updated..... 06 APR 2025
''
'' =================================================================================================


con

  VERSION = 0_2_0


con { timing }

  _xinfreq = 5_000_000                                          ' use 5MHz crystal
  _clkmode = xtal1 + pll16x                                     ' x16 = 80MHz

  CLK_FREQ = (_clkmode >> 6) * _xinfreq                         ' system freq as a constant
  MS_001   = CLK_FREQ / 1_000                                   ' ticks in 1ms
  US_001   = CLK_FREQ / 1_000_000                               ' ticks in 1us


con { terminal }

  BR_TERM  = 115_200                                            ' terminal baud rate


con { fixed io pins }

  RX_PGM   = 31  { I }                                          ' serial / programming
  TX_PGM   = 30  { O }

  SDA1     = 29  { I/O }                                        ' i2c / eeprom
  SCL1     = 28  { I/O }

  LED27    = 27  { O }                                          ' LEDs on PAB and FLiP
  LED26    = 26  { O }


con { app io pins }

  RTC_SDA  = 17  { I/O }
  RTC_SCL  = 16  { I/O }


con

  #true,  ON, OFF
  #false, NO, YES


obj

' main                                                          ' * master Spin cog
  time : "jm_time_80"                                           '   timing and delays (80MHz)
  io   : "jm_io"                                                '   simple io
  term : "jm_fullduplexserial"                                  ' * serial IO for terminal
  rtc  : "jm_ds3231"

' * uses cog when loaded


dat { pre-initialized }


var { globals }

  long  lasts
  long  lastd

' RTC registers
' -- these are BCD

  byte  seconds
  byte  minutes
  byte  hours
  byte  day
  byte  date
  byte  month
  byte  year

  byte  display                                                 ' time display type


pub main

  setup
  wait_for_terminal(true, 250)

  check_clock_set

  lasts := -1                                                   ' force initial updates
  lastd := -1

  repeat
    show_time_date


pub check_clock_set : result | t, count, check

'' Press [Y] during countdown to reset RTC registers
'' -- and other key aborts countdown

  term.fstr0(string("Press [Y] key to update RTC\r--"))

  t := cnt                                                      ' start timer

  repeat count from 3 to 1                                      ' 3s countdown
    term.fstr1(string(" %d"), count)
    repeat 4
      check := term.rxcheck
      if (check == "y") or (check == "Y")
        update_rtc
        return
      elseif (check <> -1)                                      ' other key?
        return                                                  '  yes, abort
      term.tx(".")
      waitcnt(t += constant(250 * MS_001))                      ' 4 x 250ms = 1s


dat { prompt strings }

  s_Year        byte    "Year (0..99).............. ", 0
  s_Month       byte    "Month (1..12)............. ", 0
  s_Date        byte    "Date (1..31).............. ", 0        ' modified by code
  s_Day         byte    "Day (1 [Sun]..7 [Sat]).... ", 0
  s_Display     byte    "Time display (12 or 24)... ", 0
  s_AmPm        byte    "AM or PM.................. ", 0
  s_Hours12     byte    "Hours (1..12)............. ", 0
  s_Hours24     byte    "Hours (0..23)............. ", 0
  s_Minutes     byte    "Minutes (0..59)........... ", 0
  s_Seconds     byte    "Seconds (0..59)........... ", 0


pub update_rtc | ampm, md

'' Prompt and get RTC values

  term.fstr0(string("\016Update RTC Registers\r\r"))

  year  := rtc.dec2bcd(prompted_value(@s_Year,  0, 99, "-"))
  month := rtc.dec2bcd(prompted_value(@s_Month, 1, 12, "-"))

  md := rtc.last_date(month, year)                              ' get last day of month
  s_Date[09] := (md  / 10) + "0"                                ' update prompt
  s_Date[10] := (md // 10) + "0"

  date := rtc.dec2bcd(prompted_value(@s_Date,  1, md, "-"))
  day  := rtc.dec2bcd(prompted_value(@s_Day,   1,  7, "-"))

  display := prompted_value(@s_Display, 12, 24, "|")            ' 12 or 24

  if (display == 12)
    hours := prompted_value(@s_Hours12, 1, 12, "-")
    repeat
      term.str(@s_AmPm)
      ampm := term.rx
      if (ampm := lookdown(ampm : "aApP"))                      ' good input?
        quit                                                    '  yes, move along
      else                                                      ' no
        term.tx(13)                                             '  prompt again
    ampm := --ampm / 2                                          ' AM = 0, PM = 1
    term.tx("A" + (ampm * 15))
    term.tx(13)
    if (ampm == 0)
      if (hours == 12)
        hours := 0                                              ' correct 12AM
    elseif (ampm == 1)
      if (hours < 12)
        hours := hours + 12                                     ' add pm offset
    hours := rtc.dec2bcd(hours)

  else
    hours := rtc.dec2bcd(prompted_value(@s_Hours24, 0, 23, "-"))

  minutes := rtc.dec2bcd(prompted_value(@s_Minutes, 0, 59, "-"))
  seconds := rtc.dec2bcd(prompted_value(@s_Seconds, 0, 59, "-"))

  rtc.write(7, rtc#R_SCS, @seconds)                             ' update RTC

  term.tx(term#CLS)


pub prompted_value(s_prompt, minval, maxval, mode) : result

  repeat
    term.str(s_prompt)
    result := get_dec
    if (mode == "|")                                            ' min or max
      if (result == minval) or (result == maxval)
        return
    else                                                        ' in range, min to max
      if (result => minval) and (result =< maxval)
        return


pub get_dec : value | c

'' Simple decimal input
'' -- use CR to finish
'' -- disable PST echo

  repeat
    c := term.rx
    case c
      "0".."9" :                                                ' decimal digit?
        term.tx(c)
        value := (value * 10) + (c - "0")

      13 :                                                      ' CR?
        term.tx(13)
        return value

      other :                                                   ' abort on bad character
        term.tx(13)
        return 0


pub show_time_date | check, d, m, y

  rtc.read(7, @seconds, rtc#R_SCS)                              ' read RTC registers

  if (seconds == lasts)                                         ' if no seconds change
    return

  term.tx(term#HOME)

  term.fstr1(string("%s\011\r"), rtc.day_long(day))             ' show long day name, CLR_EOL (\011)

  term.fstr3(string("%s %x, 20%.2x\011\r"), {                   ' show date with long month name
  } rtc.month_long(month), date, year)

  if (display == 24)
    term.fstr1(string("%s\r\r"), rtc.time_hhmmss(@seconds))
  else
    term.fstr1(string("%s\r\r"), rtc.time_ampm(@seconds))

  lasts := seconds                                              ' update on next second


  if (date == lastd)                                            ' new day?
    return

  term.fstr1(string("Ordinal Date... %d\011\r"), {
  } rtc.ordinal(date, month, year))

  d := rtc.bcd2dec(date)                                        ' convert for Julian Day
  m := rtc.bcd2dec(month)
  y := rtc.bcd2dec(year) + 2000

  term.fstr1(string("Julian Day..... %d\011\r"), {
  } rtc.julian(d, m, y))

  lastd := date                                                 ' update on day change


pub setup

'' Setup IO and objects for application

  time.start                                                    ' setup timing & delays

  io.start(0, 0)                                                ' clear all pins (master cog)

  term.tstart(BR_TERM)                                          ' start serial for terminal *

  rtc.start(RTC_SCL, RTC_SDA)                                   ' connect to RTC


pub wait_for_terminal(clear, delay)

'' Wait for terminal to be open and key pressed

  term.rxflush
  term.rx
  if (clear)
    term.tx(term#CLS)
    time.pause(delay #> 0)


con { license }

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}